import json
import uuid
from datetime import datetime

import requests
from requests.exceptions import RequestException
from typing_extensions import Literal
from typing import Any, Dict, List, Optional, Union
from enum import Enum
from uuid import UUID, uuid4
from .data.groundtruth import validate_groundtruth_entry
from .data.requester_question_example import validate_requester_example_image
from .data.requester_restricted_answer_set import validate_requester_restricted_answer_set_uris
from .data.taskdata import validate_taskdata_entry, Entity
from pydantic.v1 import BaseModel, validator, ValidationError, validate_model, HttpUrl, AnyHttpUrl, root_validator
from pydantic.v1.error_wrappers import ErrorWrapper
from pydantic.v1.fields import Field
from decimal import Decimal
from basemodels.manifest.restricted_audience import RestrictedAudience
from basemodels.constants import JOB_TYPES_FOR_CONTENT_TYPE_VALIDATION, BaseJobTypesEnum


# A validator function for UUID fields
def validate_uuid(cls, value):
    return value or uuid.uuid4()


# Return a request type validator function
class RequestTypeValidator(object):
    def __init__(self, multi_challenge: bool = True):
        self.multi_challenge = multi_challenge

    def validate(self, cls, value, values):
        """
        validate request types for all types of challenges
        multi_challenge should always have multi_challenge_manifests
        """
        if value == BaseJobTypesEnum.multi_challenge:
            if not self.multi_challenge:
                raise ValidationError("multi_challenge request is not allowed here.")
            if "multi_challenge_manifests" not in values:
                raise ValidationError("multi_challenge requires multi_challenge_manifests.")
        elif value in [BaseJobTypesEnum.image_label_multiple_choice, BaseJobTypesEnum.image_label_area_select]:
            if values.get("multiple_choice_min_choices", 1) > values.get("multiple_choice_max_choices", 1):
                raise ValidationError("multiple_choice_min_choices cannot be greater than multiple_choice_max_choices")
        return value


# Shape types enum
class ShapeTypes(str, Enum):
    point = "point"
    bounding_box = "bounding_box"
    polygon = "polygon"


class Model(BaseModel):
    def to_primitive(self):
        return self.dict()

    # Helper function for using in the unittest
    def check(self, return_new=False):
        self.__class__.validate(self.dict())
        out_dict, _, validation_error = validate_model(self.__class__, self.__dict__)
        if validation_error:
            raise validation_error

        # For compatibility with tests
        if return_new:
            return self.__class__(**out_dict)


class Webhook(Model):
    """Model for webhook configuration"""

    webhook_id: UUID
    chunk_completed: Optional[List[str]]
    job_completed: Optional[List[str]]

    # States that might be interesting later
    # job_skipped: List[str] = None
    # job_inserted : List[str] = None
    # job_activated : List[str] = None


class TaskData(BaseModel):
    """objects within taskdata list in Manifest"""

    task_key: UUID
    datapoint_uri: Optional[AnyHttpUrl]
    entities: Optional[List[Entity]]
    polygon: Optional[List[int]]
    datapoint_text: Optional[Dict[str, str]]
    datapoint_hash: Optional[str] = Field(..., min_length=10, strip_whitespace=True)

    @validator("datapoint_uri", always=True)
    def validate_datapoint_uri(cls, value):
        if value and len(value) < 10:
            raise ValueError("datapoint_uri need to be at least 10 char length.")
        return value

    @root_validator
    def validate_task_data(cls, values):
        """
        Validate datapoint_uri.

        Raise error if no datapoint_text and no value for URI.
        """
        if not values.get("datapoint_uri") and not values.get("datapoint_text"):
            raise ValueError("datapoint_uri is missing.")
        return values


class RequestConfig(Model):
    """definition of the request_config object in manifest"""

    version: int = 0
    shape_type: Optional[ShapeTypes]
    min_points: Optional[int]
    max_points: Optional[int]
    min_shapes_per_image: Optional[int]
    max_shapes_per_image: Optional[int]
    restrict_to_coords: Optional[bool]
    minimum_selection_area_per_shape: Optional[int]
    multiple_choice_max_choices: Optional[int] = 1
    multiple_choice_min_choices: Optional[int] = 1
    overlap_threshold: Optional[float]
    answer_type: Optional[str] = "str"
    max_value: Optional[float]
    min_value: Optional[float]
    max_length: Optional[int]
    min_length: Optional[int]
    sig_figs: Optional[int]
    keep_answers_order: Optional[bool]
    ignore_case: Optional[bool] = False
    enable_hold_time: Optional[bool] = False


class InternalConfig(Model):
    """discarded from incoming manifests"""

    exchange: Optional[Dict[str, Union[str, int, float]]]
    reco: Optional[Dict[str, Union[str, int, float]]]
    repo: Optional[Dict[str, Union[str, int, float]]]
    other: Optional[Dict[str, Union[str, int, float]]]
    # Accept one layer of nested
    mitl: Optional[Dict[str, Union[str, int, float, Dict[str, Union[str, int, float]]]]]

    class Config:
        arbitrary_types_allowed = True


class NestedManifest(Model):
    """The nested manifest description for multi_challenge jobs"""

    # We will set a default dynamic value for job_id
    job_id: Optional[UUID]
    validate_job_id = validator("job_id", always=True, allow_reuse=True)(validate_uuid)

    request_type: BaseJobTypesEnum
    validate_request_type = validator("request_type", allow_reuse=True, always=True)(
        RequestTypeValidator(multi_challenge=False).validate
    )
    requester_restricted_answer_set: Optional[Dict[str, Dict[str, str]]]

    @validator("requester_restricted_answer_set", always=True)
    def validate_requester_restricted_answer_set(cls, value, values, **kwargs):
        """image_label_area_select should always have a single RAS set"""

        # validation runs before other params, so need to handle missing case
        if "request_type" not in values:
            raise ValueError("request_type missing")
        if values["request_type"] == BaseJobTypesEnum.image_label_area_select:
            if not value or len(value.keys()) == 0:
                value = {"label": {}}
                values["requester_restricted_answer_set"] = value
        if values["request_type"] == BaseJobTypesEnum.image_label_multiple_choice:
            if not value or len(value.keys()) <= 1:
                raise ValueError(
                    "image_label_multiple_choice needs at least 2+ options in requester_restricted_answer_set"
                )
            elif len(value.keys()) > 4:
                raise ValueError(
                    "image_label_multiple_choice can not handle more than 4 options requester_restricted_answer_set"
                )
        return value

    requester_description: Optional[str]
    requester_max_repeats: int = 100
    requester_min_repeats: int = 1
    requester_question: Optional[Dict[str, str]]
    requester_question_example: Optional[Union[HttpUrl, List[HttpUrl]]]

    @validator("requester_question_example")
    def validate_requester_question_example(cls, value, values, **kwargs):
        # validation runs before other params, so need to handle missing case
        if not ("request_type" in values):
            raise ValueError("request_type missing")

        # based on https://github.com/hCaptcha/hmt-basemodels/issues/27#issuecomment-590706643
        supports_lists = [BaseJobTypesEnum.image_label_area_select, BaseJobTypesEnum.image_label_binary]

        if isinstance(value, list) and not values["request_type"] in supports_lists:
            raise ValueError("Lists are not allowed in this challenge type")
        return value

    unsafe_content: bool = False
    requester_accuracy_target: float = 0.1

    request_config: Optional[RequestConfig]

    # Groundtruth data is stored as a URL or optionally as an inlined json-serialized stringtype
    groundtruth_uri: Optional[HttpUrl]
    groundtruth: Optional[str]

    @validator("groundtruth", always=True)
    def validate_groundtruth(cls, v, values, **kwargs):
        if "groundtruth_uri" in values and "groundtruth" in values:
            raise ValueError("Specify only groundtruth_uri or groundtruth, not both.")
        return v

    # Configuration id -- XXX LEGACY
    confcalc_configuration_id: Optional[str]
    webhook: Optional[Webhook]

    class Config:
        arbitrary_types_allowed = True


class Manifest(Model):
    """The manifest description."""

    job_mode: Literal["batch", "online", "instant_delivery"]

    # We will set a default dynamic value for job_api_key
    job_api_key: Optional[UUID]

    # We will set a default dynamic value for job_id
    job_id: UUID = Field(default_factory=uuid4)

    job_total_tasks: int
    multi_challenge_manifests: Optional[List[NestedManifest]]
    request_type: BaseJobTypesEnum
    network: Optional[str]
    only_sign_results: bool = False
    public_results: bool = False

    requester_restricted_answer_set: Optional[Dict[str, Dict[str, str]]]

    requester_description: Optional[str]
    requester_max_repeats: int = 100
    requester_min_repeats: int = 1
    requester_question: Optional[Dict[str, str]]

    requester_question_example: Optional[Union[HttpUrl, List[HttpUrl]]]
    requester_example_extra_fields: Optional[Union[Dict[str, str], List[Union[Dict[str, str]]]]]

    unsafe_content: bool = False
    task_bid_price: float
    oracle_stake: Decimal
    expiration_date: Optional[int]
    start_date: Optional[int]
    requester_accuracy_target: float = 0.1
    manifest_smart_bounty_addr: Optional[str]
    hmtoken_addr: Optional[str]
    minimum_trust_server: float = 0.1
    minimum_trust_client: float = 0.1
    recording_oracle_addr: Optional[str]
    reputation_oracle_addr: Optional[str]
    reputation_agent_addr: Optional[str]
    requester_pgp_public_key: Optional[str]
    ro_uri: Optional[str]
    repo_uri: Optional[str]
    batch_result_delivery_webhook: Optional[AnyHttpUrl]
    online_result_delivery_webhook: Optional[AnyHttpUrl]
    instant_result_delivery_webhook: Optional[AnyHttpUrl]

    request_config: Optional[RequestConfig]

    # If taskdata is directly provided
    taskdata: Optional[List[TaskData]]

    # If taskdata is separately stored
    taskdata_uri: Optional[AnyHttpUrl]

    # Groundtruth data is stored as a URL or optionally as an inlined json-serialized stringtype
    groundtruth_uri: Optional[AnyHttpUrl]
    groundtruth: Optional[str]

    # internal config options for param tests etc.
    internal_config: Optional[InternalConfig]

    # Configuration id
    confcalc_configuration_id: Optional[str]
    restricted_audience: Optional[RestrictedAudience] = {}

    webhook: Optional[Webhook]

    rejected_uri: Optional[AnyHttpUrl]
    rejected_count: Optional[int]

    is_verification: bool = False

    # #### Validators

    @root_validator
    def validate(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        start_date = values["start_date"]
        expiration_date = values["expiration_date"]
        # validate at least taskdata or taskdata_uri is present
        taskdata = values.get("taskdata")
        taskdata_uri = values.get("taskdata_uri")
        if taskdata is not None and len(taskdata) > 0 and taskdata_uri is not None:
            raise ValueError("Specify only one of taskdata {} or taskdata_uri {}".format(taskdata, taskdata_uri))
        if taskdata is None and taskdata_uri is None:
            raise ValueError("No taskdata or taskdata_uri found in manifest")

        if not start_date and not expiration_date:
            # Timestamps are not passed
            return values
        has_both_dates = bool(start_date) and bool(expiration_date)
        if not has_both_dates:
            raise ValueError("You must specify both start_date and expiration_date")

        if not start_date < expiration_date:
            raise ValueError("start_date must be before expiration_date")

        duration = datetime.utcfromtimestamp(expiration_date) - datetime.utcfromtimestamp(start_date)
        if 7 < duration.days:
            raise ValueError("Max job duration is 7 days.")

        return values

    @validator("requester_min_repeats")
    def validate_min_repeats(cls, v, values):
        """min repeats are required to be at least 4 if ilmc"""
        if values["request_type"] == "image_label_multiple_choice":
            return max(v, 4)
        return v

    @validator("groundtruth", always=True)
    def validate_groundtruth(cls, value, values):
        if "groundtruth_uri" in values and "groundtruth" in values:
            raise ValueError("Specify only groundtruth_uri or groundtruth, not both.")
        return value

    @validator("requester_restricted_answer_set", always=True)
    def validate_requester_restricted_answer_set(cls, value, values, **kwargs):
        """image_label_area_select should always have a single RAS set"""
        # validation runs before other params, so need to handle missing case
        if not ("request_type" in values):
            raise ValueError("request_type missing")

        if values["request_type"] == BaseJobTypesEnum.image_label_area_select:
            if not value or len(value.keys()) == 0:
                value = {"label": {}}
                values["requester_restricted_answer_set"] = value

        if values["request_type"] == BaseJobTypesEnum.image_label_multiple_choice:
            if not value or len(value.keys()) <= 1:
                raise ValueError(
                    "image_label_multiple_choice needs at least 2+ options in requester_restricted_answer_set"
                )
            elif len(value.keys()) > 4:
                raise ValueError(
                    "image_label_multiple_choice can not handle more than 4 options requester_restricted_answer_set"
                )
        return value

    @validator("requester_question_example")
    def validate_requester_question_example(cls, value, values, **kwargs):
        # validation runs before other params, so need to handle missing case
        if not ("request_type" in values):
            raise ValueError("request_type missing")

        # based on https://github.com/hCaptcha/hmt-basemodels/issues/27#issuecomment-590706643
        supports_lists = [
            BaseJobTypesEnum.image_label_area_select,
            BaseJobTypesEnum.image_label_binary,
        ]

        if isinstance(value, list) and not (values["request_type"] in supports_lists):
            raise ValueError("Lists are not allowed in this challenge type")
        return value

    validate_request_type = validator("request_type", allow_reuse=True, always=True)(RequestTypeValidator().validate)
    validate_job_api_key = validator("job_api_key", always=True, allow_reuse=True)(validate_uuid)
    validate_job_id = validator("job_id", always=True, allow_reuse=True)(validate_uuid)

    class Config:
        arbitrary_types_allowed = True

    def to_primitive(self):
        """Override primitive function to make it serializable."""
        manifest_json = self.json()
        return json.loads(manifest_json)


def validate_groundtruth_uri(manifest: dict):
    """
    Validate groundtruth_uri
    Returns entries count if succeeded
    """
    request_type = manifest.get("request_type", "")
    validate_image_content_type = request_type in JOB_TYPES_FOR_CONTENT_TYPE_VALIDATION
    uri_key = "groundtruth_uri"
    uri = manifest.get(uri_key)
    if uri is None:
        return
    try:
        response = requests.get(uri, timeout=(3.5, 5))
        response.raise_for_status()

        entries_count = 0
        data = response.json()
        if isinstance(data, dict):
            for k, v in data.items():
                entries_count += 1
                validate_groundtruth_entry(k, v, request_type, validate_image_content_type)
                validate_image_content_type = False
        else:
            for v in data:
                entries_count += 1
                validate_groundtruth_entry("", v, request_type, validate_image_content_type)
                validate_image_content_type = False

    except (ValidationError, RequestException) as e:
        raise ValidationError(
            [ErrorWrapper(ValueError(f"Validation failed for {uri}: {e}"), uri_key)], Manifest
        ) from e

    if entries_count == 0:
        raise ValidationError([ErrorWrapper(ValueError(f"fetched {uri} is empty"), uri_key)], Manifest)


def validate_taskdata_uri(manifest: dict):
    """
    Validate taskdata_uri
    Returns entries count if succeeded
    """
    request_type = manifest.get("request_type", "")
    validate_image_content_type = request_type in JOB_TYPES_FOR_CONTENT_TYPE_VALIDATION
    uri_key = "taskdata_uri"
    uri = manifest.get(uri_key)
    if uri is None:
        return
    try:
        response = requests.get(uri, timeout=(3.5, 5))
        response.raise_for_status()

        entries_count = 0
        data = response.json()
        for v in data:
            entries_count += 1
            validate_taskdata_entry(v, validate_image_content_type)
            validate_image_content_type = False  # We want to validate only first entry for content type

    except (ValidationError, RequestException) as e:
        raise ValidationError(
            [ErrorWrapper(ValueError(f"Validation failed for {uri}: {e}"), uri_key)], Manifest
        ) from e

    if entries_count == 0:
        raise ValidationError([ErrorWrapper(ValueError(f"fetched {uri} is empty"), uri_key)], Manifest)


def validate_manifest_example_images(manifest: dict):
    """Fetch and validate the example resources."""
    question_example = manifest.get("requester_question_example")
    req_res_answer_set = manifest.get("requester_restricted_answer_set", {})
    if question_example:
        # some jobs might not have requester_question_example
        validate_requester_example_image(question_example)
    if req_res_answer_set:
        validate_requester_restricted_answer_set_uris(req_res_answer_set)


def validate_manifest_uris(manifest: dict):
    """Fetch & validate manifest's remote objects"""
    validate_taskdata_uri(manifest)
    validate_groundtruth_uri(manifest)
